#
# Top Level Classes
#


class Status:
    pass


class Untested(Status):
    pass


class Tested(Status):
    pass


class Success(Tested):
    pass


class Failed(Tested):
    requires_exception: bool = False
    reason: str = ""  # Set by subclasses

    @classmethod
    def kind(cls):
        if cls.requires_exception:
            return cls.__name__
        return f"{cls.__name__}: {cls.reason}"


#
# Failure Classes
#


# Do not open URLs


class NoOpen(Failed):
    pass


class Skipped(NoOpen):
    reason = "This domain was skipped"


class Unknown(NoOpen):
    reason = "This website is for an unknown / unsupported domain."


class BadRequest(NoOpen):
    requires_exception = True
    reason = "Request failed"

    def __init__(self, why: Exception):
        self.exc: Exception = why


class MalformedURL(NoOpen):
    pass


class NotInt(MalformedURL):
    reason = "The URL's chapter is not an integer."


class HasVol(MalformedURL):
    reason = "The URL contains 'vol'; this is a bad sign"


class Pattern(MalformedURL):
    reason = "The URL contains a pattern that is dangerous"


# Open URLs


class ToOpen(Failed):
    pass


class Tiny(ToOpen):
    reason = "URL failed by default, it is too small"


class Exists(ToOpen):
    reason = "The URL is valid, but this site seems to be missing other chapters"


class Missing(ToOpen):
    reason = "Previous and future chapters exist, this one does not."


class Broken(ToOpen):
    reason = "This website does not seem to have any chapters of this manga."



import contextlib
import inspect
import tqdm


def _print(obj, **kwargs):
    tqdm.tqdm.write(obj if isinstance(obj, str) else str(obj))


@contextlib.contextmanager
def redirect_print_to_tqdm():
    old = print
    try:
        inspect.builtins.print = _print
        yield
    finally:
        inspect.builtins.print = old



from collections.abc import Callable
from random import choice
from time import sleep

import tldextract
import requests

from .unknown_domain import UnknownDomain
from .domains import domains


_agents = (
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
    "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.8.1.6) Gecko/20070802 SeaMonkey/1.1.4",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 14_4_1) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.3.1 Safari/605.1.15",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:124.0) Gecko/20100101 Firefox/124.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 14.4; rv:124.0) Gecko/20100101 Firefox/124.0",
    "Mozilla/5.0 (X11; Linux i686; rv:124.0) Gecko/20100101 Firefox/124.0",
)


def get_domain(url: str) -> str:
    """
    Extract the domain portion of a URL
    """
    info: tldextract.tldextract.ExtractResult = tldextract.extract(url)
    return f"{info.domain}.{info.suffix}"


def _test(url: str, fn: Callable[[str], bool], timeout: float, timeout_retries: int, delay: float) -> bool:
    """
    Return true if the given chapter is found
    If a timeout occurs, retries at most timeout_retries times; sleeps in between
    """
    session = requests.Session()
    session.headers.update({"User-Agent": choice(_agents)})
    what: str
    try:
        response = session.get(url, timeout=timeout)
        if response.ok:
            return fn(response.text)
        if response.status_code not in (429, 503):
            print(f"Got {response.status_code} from GET {url}")
            return False
        what = f"Got {response.status_code}"
    except requests.exceptions.ReadTimeout:
        what = "ReadTimeout"
    # Timeout or too many requests error at this point
    if timeout_retries <= 0:
        raise
    if delay > 60:
        print(f"{what} for: {url}: Sleeping for {delay} seconds then trying again")
    sleep(delay)
    return _test(url, fn, timeout, timeout_retries - 1, min(delay * 2, 960.0))


def test(url: str, timeout: float = 7.5, timeout_retries: int = 0, base_delay: float = 7.5) -> bool:
    """
    Return true if the given chapter is found
    If a timeout occurs, retries at most timeout_retries times; sleeps in between
    """
    domain: str = get_domain(url)
    try:
        fn: Callable[[str], bool] = domains[domain]
    except KeyError:
        raise UnknownDomain(url)  # pylint: disable=raise-missing-from
    return _test(url, fn, timeout, timeout_retries, base_delay)
